//=============================================================================
// Point-defence. Destroys incoming projectiles.
// http://mrevil.pwp.blueyonder.co.uk/unreal/
//=============================================================================
class SentinelUpgrade_APS extends SentinelUpgrade;

var const CylinderComponent CylinderComponent;

/** Maximum distance at which projectiles can be intercepted, relative to Sentinel's size. */
var() float MaxRange;
/** Range within which projectiles are too close to be intercepted. */
var() float MinRange;
/** Minimum time between shots. */
var() float RefireInterval;
/** Projectiles travelling faster than this cannot be intercepted. */
var() float MaximumProjectileSpeed;

/** Effect to use for beam. */
var() ParticleSystem BeamTemplate;
/** Played when beam fires. */
var() SoundCue BeamSound;

struct DestroyedProjectileInfo
{
	var Actor Projectile; //Not of type Projectile because it could be a Redeemer.
	var class<Projectile> ProjectileClass; //Replicated so that bNetTemporary projectiles can be destroyed client-side.
	var Vector ProjectileLocation; //Replicated so that bNetTemporary projectiles can be destroyed client-side.
};

/** This is replicated so that effects can be spawned on the client. */
var repnotify DestroyedProjectileInfo DestroyedProjectile;

replication
{
	if(Role == ROLE_Authority && bNetDirty && (DestroyedProjectile.ProjectileClass != none || DestroyedProjectile.Projectile != none))
		DestroyedProjectile;
}

simulated event ReplicatedEvent(name VarName)
{
	if(VarName == 'DestroyedProjectile')
	{
		ClientDestroyProjectile();
	}
	else
	{
		super.ReplicatedEvent(VarName);
	}
}

function InitializeFor(Sentinel S)
{
	super.InitializeFor(S);

	CylinderComponent.SetCylinderSize(S.CylinderComponent.CollisionRadius * MaxRange, S.CylinderComponent.CollisionHeight * MaxRange);
	CylinderComponent.SetTranslation(S.CylinderComponent.Translation);
	Activate();
}

function NotifyDied(Controller Killer, out class<DamageType> DamageType, vector HitLocation)
{
	super.NotifyDied(Killer, DamageType, HitLocation);

	Deactivate();
}

function SentinelUpgrade DetachUpgrade(SentinelUpgrade UpgradeToDetach)
{
	if(UpgradeToDetach == self)
	{
		Deactivate();
	}

	return super.DetachUpgrade(UpgradeToDetach);
}

function Touch(Actor Other, PrimitiveComponent OtherComp, Vector HitLocation, Vector HitNormal)
{
	local Projectile Proj;
	local UTRemoteRedeemer Redeemer;

	if((VSize(Other.Velocity - Velocity) > MaximumProjectileSpeed) || (VSize(Other.Location - CylinderComponent.GetPosition()) < MinRange * Cannon.CylinderComponent.CollisionRadius))
	{
		return;
	}

	Proj = Projectile(Other);

	if(Proj != none)
	{
		if(Proj.InstigatorController != Cannon.InstigatorController && !WorldInfo.GRI.OnSameTeam(Cannon, Proj))
		{
			DestroyedProjectile.Projectile = Proj;
			DestroyedProjectile.ProjectileClass = Proj.Class;
			DestroyedProjectile.ProjectileLocation = Proj.Location;

			Deactivate(RefireInterval);
			SpawnEffectsTo(Proj.Location);
			Cannon.Controller.HearNoise(1.0, Proj);
			Proj.Explode(Proj.Location, Normal(Proj.Velocity));
		}
	}
	else
	{
		Redeemer = UTRemoteRedeemer(Other);

		if(Redeemer != none)
		{
			if(Redeemer.InstigatorController != Cannon.InstigatorController && !WorldInfo.GRI.OnSameTeam(Cannon, Redeemer))
			{
				DestroyedProjectile.Projectile = Redeemer;
				DestroyedProjectile.ProjectileClass = none;
				DestroyedProjectile.ProjectileLocation = Redeemer.Location;

				Deactivate(RefireInterval);
				SpawnEffectsTo(Redeemer.Location);
				Cannon.Controller.HearNoise(1.0, Redeemer);
				Redeemer.BlowUp();
			}
		}
	}
}

/**
 * If the projectile destroyed on the server was bNetTemporary, then this searches for the closest projectile of the same class on the client and destroys it. Also spawns effects.
 */
simulated function ClientDestroyProjectile()
{
	local Projectile Proj, ClosestProj;
	local float ProjDist, ClosestProjDist;

	if(DestroyedProjectile.Projectile == none)
	{
		foreach DynamicActors(class'Projectile', Proj)
		{
			if(Proj.Class == DestroyedProjectile.ProjectileClass)
			{
				ProjDist = VSize(Proj.Location - DestroyedProjectile.ProjectileLocation);

				if(ClosestProj == none || ProjDist < ClosestProjDist)
				{
					ClosestProj = Proj;
					ClosestProjDist = ProjDist;
				}
			}
		}

		if(ClosestProj != none && ClosestProjDist < 200.0)
		{
			ClosestProj.Explode(ClosestProj.Location, Normal(ClosestProj.Velocity));
		}
	}

	SpawnEffectsTo(DestroyedProjectile.ProjectileLocation);
}

/**
 * Spawns a beam emitter and plays a sound.
 */
simulated function SpawnEffectsTo(Vector EndLocation)
{
	local ParticleSystemComponent E;
	local Vector StartLocation;

	if(EffectIsRelevant(Cannon.Location, false))
	{
		StartLocation = Cannon.GetPawnViewLocation();
		E = WorldInfo.MyEmitterPool.SpawnEmitter(BeamTemplate, StartLocation);

		if(E != none)
		{
			E.SetVectorParameter('BeamEnd', EndLocation);
		}

		PlaySound(BeamSound, true,,, StartLocation + ((EndLocation - StartLocation) * 0.5));
	}
}

/**
 * Disables collision, preventing projectiles from being intercepted.
 *
 * @param		DeactivationTime	length of time to remain deactivated. If 0, then deactivation is for an indefinite time
 */
protected simulated function Deactivate(optional float DeactivationTime)
{
	SetCollision(false, bBlockActors, bIgnoreEncroachers);

	if(DeactivationTime > 0.0)
	{
		SetTimer(DeactivationTime, false, 'Activate');
	}
}

/**
 * Enables collision, allowing projectiles to be intercepted.
 */
protected simulated function Activate()
{
	SetCollision(true, bBlockActors, bIgnoreEncroachers);
}

defaultproperties
{
	FriendlyName="Active Protection System"
	Description="The new point-defence module, recently released for field trials by authorised personnel.\n\nAvailable information indicates a recharge time of 0.3 seconds, theoretically allowing up to 200 projectiles per minute to be detected and destroyed before they reach the Sentinel. Initial reports indicate that this is sufficient to render the Sentinel effectively invulnerable to many common projectile weapons."

	MaxRange=6.5
	MinRange=3.0
	RefireInterval=0.3
	MaximumProjectileSpeed=16000.0
	BeamTemplate=ParticleSystem'SentinelGear_Resources.Effects.Upgrades.P_APS_Beam'
	BeamSound=SoundCue'SentinelGear_Resources.Sounds.Upgrades.APS_BeamSoundCue'

	AmmoCost=100
	ActivateSound=SoundCue'SentinelGear_Resources.Sounds.Upgrades.APS_Activate'

	bCollideActors=false
	bBlockActors=false

	Begin Object Class=CylinderComponent Name=CollisionCylinder
		CollisionRadius=0.0
		CollisionHeight=0.0
		BlockNonZeroExtent=true
		BlockZeroExtent=true
		BlockActors=false
		CollideActors=true
	End Object
	CollisionComponent=CollisionCylinder
	CylinderComponent=CollisionCylinder
	Components.Add(CollisionCylinder)
}